#ifndef _COMPAT_H
#define _COMPAT_H 1



#include <stdint.h>

typedef uint32_t in_addr_t;
typedef uint16_t in_port_t;

#define PIPE_INIT { ((pipe_t)-1), ((pipe_t)-1) }



#if defined(__GNUC__)

#include <unistd.h>
#include <pthread.h>

typedef pthread_mutex_t mutex_t;
#define MUTEX_INIT(x) pthread_mutex_init(&x, NULL)
#define MUTEX_LOCK(x) pthread_mutex_lock(&x)
#define MUTEX_UNLOCK(x) pthread_mutex_unlock(&x)

typedef pthread_t thread_t;
typedef void * thread_func_ret_t;
#define THREAD_FUNC_TYPE
#define THREAD_CREATE(x, y, z) pthread_create(&x, NULL, y, z)
#define THREAD_DETACH(x) pthread_detach(x)
#define THREAD_JOIN(x) pthread_join(x, NULL)

typedef int pipe_t;
#define PIPE_OPEN(x, y) pipe(&x)
#define PIPE_READ(r, x, y, z) (ssize_t)((r = read(x, y, z)))
#define PIPE_WRITE(r, x, y, z) (ssize_t)((r = write(x, y, z)))
#define PIPE_CLOSE(x) close(x)



#elif defined(_MSC_VER)

#include <stdio.h>
#include <windows.h>
#include <process.h>

typedef CRITICAL_SECTION mutex_t;
#define MUTEX_INIT(x) InitializeCriticalSection(&x)
#define MUTEX_LOCK(x) EnterCriticalSection(&x)
#define MUTEX_UNLOCK(x) LeaveCriticalSection(&x)

#define PTHREAD_MUTEX_INITIALIZER { 0 };
//#define PTHREAD_MUTEX_INITIALIZER {static mutex_t m; MUTEX_INIT(m);}

typedef HANDLE thread_t;
typedef unsigned int thread_func_ret_t;
#define THREAD_FUNC_TYPE WINAPI
#define THREAD_CREATE(x, y, z) (int)((x = (HANDLE)_beginthreadex(NULL, 0, y, z, 0, NULL)) == NULL)
#define THREAD_DETACH(x) CloseHandle(x)
#define THREAD_JOIN(x) ((WaitForSingleObject((x), INFINITE) != WAIT_OBJECT_0) || !CloseHandle(x))

typedef HANDLE pipe_t;
#define PIPE_OPEN(x, y) CreatePipe(&x, &y, NULL, 0)
#define PIPE_READ(r, x, y, z) ReadFile(x, y, z, &r, NULL)
#define PIPE_WRITE(r, x, y, z) WriteFile(x, y, z, &r, NULL)
#define PIPE_CLOSE(x) CloseHandle(x)

#define sleep(x) Sleep(x)
#define snprintf(...) _snprintf(__VA_ARGS__)

#define in_addr_t unsigned long
#define in_port_t unsigned long

#else
   #error "Unsupported compiler"
#endif

#endif /* _COMPAT_H */
