#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#include "monitor.h"
#include "so.h"
#include "me.h"
#include "compat.h"

#include <Windows.h>
#include <TlHelp32.h>

struct mon_process_entry {
   unsigned int *(WINAPI *callback)(void *);
   void		*args;
   int		flags;
   LPWSTR	name;
   struct	mon_process_entry *next;
   struct	mon_process_entry *prev;
};

struct process {
   LPWSTR name;
   struct process *next;
   struct process *prev;
};

static unsigned int * WINAPI mon_process_run(void *args);
static void mon_process_dump(void);

static int process_add(struct process **l, LPWSTR name);

static void process_free(struct process **l);
static int process_find(struct process *l, LPWSTR name);

static thread_t t;
static mutex_t m;
static struct mon_process_entry *list = NULL;
static int running = 0;

static unsigned int * WINAPI mon_process_run(void *args)
{
   struct mon_process_entry *p, *e;
   struct process *current = NULL;
   FILE *fp;
   int i;
   thread_t ct;   

   while(1) {
      MUTEX_LOCK(m);

      if(!list) {
         running = 0;
         MUTEX_UNLOCK(m);
         return NULL;
      }

      process_free(&current);

      //memset(&g, 0x00, sizeof(g));
      ////glob(SO"/proc/[0-9]cmdline", 0, NULL, &g);
      //for(i = 0; i < g.gl_pathc; i++) {
      //   if(!(fp = fopen(g.gl_pathv[i], "r"))) continue;
      //   if(fgets(buf, sizeof(buf), fp)) {
      //      if(!(name = strrchr(buf, '/'))) name = buf;
      //      if(name[0] == '/') name++;
      //      process_add(&current, name);
      //   }
      //   fclose(fp);
      //}
      //globfree(&g);

	  //create a linked list of running processes
	  if(!get_process_list(&current))
		  return NULL;

      p = list;
      while(p)
	  {
         if((!(p->flags & MON_FLAG_NEG) && process_find(current, p->name)) || ((p->flags & MON_FLAG_NEG) && !process_find(current, p->name)))
		 {
            THREAD_CREATE(ct, p->callback, p->args);
            THREAD_DETACH(ct);
            if(p == list) list = p->next;
            if(p->next) p->next->prev = p->prev;
            if(p->prev) p->prev->next = p->next;
            e = p;
            p = p->next;
            if(e->name) free(e->name);
            free(e);
         } 
		 else
		 {
			OutputDebugString(p->name);
            p = p->next;			
         }
      }

      mon_process_dump();

      MUTEX_UNLOCK(m);

      sleep(1000);
   }
}

void mon_process_start(void *args) 
{   
   MUTEX_LOCK(m);
   if (!running) if (!THREAD_CREATE(t, mon_process_run, NULL)) running = 1;
   MUTEX_UNLOCK(m);   
}

void *mon_process_addwatch(unsigned int *(WINAPI *callback)(void *), void *args, int flags, ...)
{
   struct mon_process_entry *p = NULL;
   va_list va;

   if (m.SpinCount == 0)
	 MUTEX_INIT(m);

   MUTEX_LOCK(m);

   do {
      if(!list) {
         if(!(list = malloc(sizeof(struct mon_process_entry)))) break;
         list->prev = NULL;
         p = list;
      } else {
         for(p = list; p->next; p = p->next);
         if(!(p->next = malloc(sizeof(struct mon_process_entry)))) break;
         p->next->prev = p;
         p = p->next;
      }
      p->next = NULL;
      p->callback = callback;
      p->args = args;
      p->flags = flags;

      va_start(va, flags);
      p->name = _wcsdup(va_arg(va, LPWSTR));
      va_end(va);

	  if (!running) if (!THREAD_CREATE(t, mon_process_run, NULL)) running = 1;
   } while(0);

   mon_process_dump();

   MUTEX_UNLOCK(m);

   return p;
}

void mon_process_delwatch(void *entry)
{
   struct mon_process_entry *p = NULL, *e = (struct mon_process_entry *)entry;

   if(!list || !e) return;

   MUTEX_LOCK(m);

   for(p = list; p && (p != e); p = p->next);
   if(p) {
      if(p == list) list = p->next;
      if(p->next) p->next->prev = p->prev;
      if(p->prev) p->prev->next = p->next;
      if(p->name) free(p->name);
      free(p);
   }

   MUTEX_UNLOCK(m);

   return;
}

void mon_process_dump(void)
{
#if 0
   struct mon_process_entry *p = NULL;

   debugme("\n--- PROCESS MONITOR LIST ---\n");
   for(p = list; p; p = p->next) debugme("- %s (at %p)\n", p->name, (void *)p);
   debugme("\n");
#endif

   return;
}

static int process_add(struct process **l, LPWSTR name)
{
   struct process *p;

   if(!*l) {
      if(!(*l = malloc(sizeof(struct process)))) return -1;
      (*l)->prev = NULL;
      p = *l;
   } else {
	   for (p = *l; p->next; p = p->next)
	   {
		   OutputDebugString(p->name);
		   OutputDebugString("\n");
		   OutputDebugString(name);
		   OutputDebugString("\n");
		   if (!wcscmp(p->name, name))
			   return 0;
	   }
	  if (!wcsicmp(p->name, name)) return 0;
      if(!(p->next = malloc(sizeof(struct process)))) return -1;
      p->next->prev = p;
      p = p->next;
   }
   p->next = NULL;
   p->name = _wcsdup(name);

   return 1;
}


static void process_free(struct process **l)
{
   struct process *p;

   while(*l) {
      p = (*l)->next;
      if((*l)->name) free((*l)->name);
      free(*l);
      *l = p;
   }

   return;
}

static int process_find(struct process *l, LPWSTR name)
{
   
   for(; l; l = l->next) {
      //if(!fnmatch(name, l->name, FNM_PERIOD|FNM_CASEFOLD)) return 1;
	   OutputDebugString(l->name);
	   OutputDebugString("\n");
	   if (!wcsicmp(name, l->name)) 
		   return 1;
   }

   return 0;
}


//create a linked list of type 'process' the running processes
int get_process_list(struct process **p)
{
	HANDLE hProcessSnap;
	HANDLE hProcess;
	PROCESSENTRY32 pe32;
	DWORD dwPriorityClass;

	// take a snapshot of all the running processes
	hProcessSnap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
	if(hProcessSnap == INVALID_HANDLE_VALUE)
	{
		return(FALSE);
	}

	// set the size of the structure
	pe32.dwSize = sizeof(PROCESSENTRY32);

	// get first process info
	if(!Process32First(hProcessSnap, &pe32))
	{
		CloseHandle(hProcessSnap);
		return(FALSE);
	}

	// get info of all the running processes
	do
	{
		process_add(p, pe32.szExeFile);
		
	}while (Process32Next(hProcessSnap, &pe32));

	

	CloseHandle(hProcessSnap);

	return(TRUE);
}